//////////////////////////////////////////////////////////////////////////////
//
//  Xilinx, Inc. 2002                 www.xilinx.com
//
//  XAPP 291 - Self Addressing FIFO
//
//////////////////////////////////////////////////////////////////////////////
//
//  File name :       saf_1to1.v
//
//  Description :     Top-level for 1:1 SAF
//		      addrsize used to define the fifo depth(ie number of address bits),
//		      minimum 4, maximum 9 with ram in 512x36 mode
// 		      datasize defines the input and output data path width, minimum 1, maximum 32
// 		      addrsize + datasize must be less than or equal 36
// 		      This example uses addrsize = 9, and datasize = 16
// 		      flag(0) high when fifo between empty and one half full
// 		      flag(1) high when fifo between one quarter and three quarters full
// 		      flag(2) high when fifo between one half full and full
//
//  Date - revision : December 11th 2002 - v 1.2
//
//  Author :          NJS
//
//  Contact : e-mail  hotline@xilinx.com
//            phone   + 1 800 255 7778
//
//  Disclaimer: LIMITED WARRANTY AND DISCLAMER. These designs are
//              provided to you "as is". Xilinx and its licensors make and you
//              receive no warranties or conditions, express, implied,
//              statutory or otherwise, and Xilinx specifically disclaims any
//              implied warranties of merchantability, non-infringement,or
//              fitness for a particular purpose. Xilinx does not warrant that
//              the functions contained in these designs will meet your
//              requirements, or that the operation of these designs will be
//              uninterrupted or error free, or that defects in the Designs
//              will be corrected. Furthermore, Xilinx does not warrantor
//              make any representations regarding use or the results of the
//              use of the designs in terms of correctness, accuracy,
//              reliability, or otherwise.
//
//              LIMITATION OF LIABILITY. In no event will Xilinx or its
//              licensors be liable for any loss of data, lost profits,cost
//              or procurement of substitute goods or services, or for any
//              special, incidental, consequential, or indirect damages
//              arising from the use or operation of the designs or
//              accompanying documentation, however caused and on any theory
//              of liability. This limitation will apply even if Xilinx
//              has been advised of the possibility of such damage. This
//              limitation shall apply not-withstanding the failure of the
//              essential purpose of any limited remedies herein.
//
//  Copyright  2002 Xilinx, Inc.
//  All rights reserved
//
//////////////////////////////////////////////////////////////////////////////
//
`timescale 1 ps / 1ps

module saf_1to1 (
	wr_clk,					// write clock input
	rd_clk,					// read clock input
	datain,					// data input
	rst,					// reset input
	we,					// write enable input
	oe,					// output enable input
	flag,					// flag outputs
	dataout);				// output data

input 		wr_clk, rd_clk, rst, we, oe ;
input 	[15:0] 	datain;

output 	[2:0] 	flag;
output 	[15:0] 	dataout;

wire 		high ;
wire 		low ;
wire 		oe_int ;
wire 		we_int ;
wire 	[0:0] 	low1 ;
wire 	[1:0] 	low2 ;
wire 	[3:0] 	low4 ;
wire 	[7:0] 	low8 ;
wire 	[31:0] 	low32 ;
wire 	[35:0] 	douta ;
wire 	[35:0] 	doutb ;
reg 	[35:0] 	dina ;
wire 	[31:0] 	ram_douta ;
wire 	[31:0] 	ram_doutb  ;
wire 	[3:0] 	ram_pouta ;
wire 	[3:0] 	ram_poutb ;
wire 	[3:0] 	ram_pina ;
wire 	[31:0] 	ram_dina ;
reg 	[8:0] 	ram_addra ;
reg 	[8:0] 	ram_addrb  ;
wire 	[8:0] 	next_wr_address ;

assign low = 1'b0 ;
assign low1 = 1'b1 ;
assign low2 = 2'b00 ;
assign low4 = 4'b0000 ;
assign low8 = 8'b00000000 ;
assign low32 = 32'b00000000000000000000000000000000 ;
assign high = 1'b1 ;
assign oe_int = oe | rst ; //needed to make sure the ram enable is high during reset (because its a synchronous reset)
assign we_int = we | rst ; //needed to make sure the ram enable is high during reset (because its a synchronous reset)

assign douta = {ram_pouta, ram_douta} ;
assign doutb = {ram_poutb, ram_doutb} ;

always @ (next_wr_address or datain)
begin
dina[10:0] = 11'b00000000000 ;
dina[35:11] = {next_wr_address, datain} ;
end

assign ram_pina = dina[35:32] ;
assign ram_dina = dina[31:0] ;

always @ (douta)
begin
ram_addra[8:0] = douta[35:27] ;
end

always @ (doutb)
begin
ram_addrb[8:0] = doutb[35:27] ;
end

assign dataout = ram_doutb[26:11] ;

assign flag[2] =(((ram_poutb[3:2] == 2'b00) && (ram_pouta[3:2] == 2'b11)) |
		 ((ram_poutb[3:2] == 2'b01) && (ram_pouta[3:2] == 2'b10)) |
		 ((ram_poutb[3:2] == 2'b11) && (ram_pouta[3:2] == 2'b00)) |
		 ((ram_poutb[3:2] == 2'b10) && (ram_pouta[3:2] == 2'b01)))
		 ? 1'b1 : 1'b0;

assign flag[1] =(((ram_poutb[3:2] == 2'b00) && (ram_pouta[3:2] == 2'b01)) |
		 ((ram_poutb[3:2] == 2'b01) && (ram_pouta[3:2] == 2'b11)) |
		 ((ram_poutb[3:2] == 2'b11) && (ram_pouta[3:2] == 2'b10)) |
		 ((ram_poutb[3:2] == 2'b10) && (ram_pouta[3:2] == 2'b00)))
		 ? 1'b1 : 1'b0;

assign flag[0] =(((ram_poutb[3:2] == 2'b00) && (ram_pouta[3:2] == 2'b00)) |
		 ((ram_poutb[3:2] == 2'b01) && (ram_pouta[3:2] == 2'b01)) |
		 ((ram_poutb[3:2] == 2'b11) && (ram_pouta[3:2] == 2'b11)) |
		 ((ram_poutb[3:2] == 2'b10) && (ram_pouta[3:2] == 2'b10)))
		 ? 1'b1 : 1'b0;

RAMB16_S36_S36 block_ram (
	.DIA(ram_dina),
	.DIB(low32),
	.DIPA(ram_pina),
	.DIPB(low4),
	.ENA(we_int),
	.ENB(oe_int),
	.WEA(we),
	.WEB(low),
	.SSRA(rst),
	.SSRB(rst),
	.CLKA(wr_clk),
	.CLKB(rd_clk),
	.ADDRA(ram_addra),
	.ADDRB(ram_addrb),
	.DOA(ram_douta),
	.DOB(ram_doutb),
	.DOPA(ram_pouta),
	.DOPB(ram_poutb));

incrementer_9bit inc (
	.ip(douta[35:27]),
	.op(next_wr_address));

endmodule

